"""Top-level package for L3 Auto-segmentation Tool."""

__author__ = """Ralph Brecheisen"""
__email__ = 'ralph.brecheisen@gmail.com'
__version__ = '0.8.0'

from .train.train import TrainL3
from . import models
from . import utils
